module SimplePayment
  class Payment < ActiveRecord::Base
    include AASM

    self.table_name = 'simple_payments'

    belongs_to :payable, polymorphic: true
    has_many :transactions

    validates_presence_of :amount, :payable_id, :payable_type, :status

    scope :accepted, ->{ where status: %w(approved completed) }

    aasm column: 'status' do
      state :pending
      state :approved
      state :cancelled
      state :completed
      state :failed
      state :refunded

      event :_approve do
        transitions from: :pending, to: :approved
      end

      event :_fail do
        transitions from: :pending, to: :failed
      end

      event :_complete do
        transitions from: :approved, to: :completed
      end

      event :_cancel do
        transitions from: :approved, to: :cancelled
      end

      event :_refund do
        transitions from: :completed, to: :refunded
      end
    end

    # Calls the gateway to execute the payment, saving the
    # communication details and status change to the model
    def execute!(nonce)
      return unless may__approve?
      request, response = gateway.execute(self, nonce)
      if response.success?
        _approve
      else
        _fail
      end
      self.external_id ||= response.external_id
      self.external_fee = gateway.calculate_fee(self)
      save!

      persist_transaction request, response
      response.success?
    rescue TransmissionError => e
      handle_error 'execute', e
      false
    end

    # Calls the gateway to see if an approved payment has been completed
    def complete!
      return unless may__complete?
      _complete! if gateway.completed?(self)

    rescue TransmissionError => e
      handle_error 'complete', e
    end

    # Cancel a payment that is approved, but not settled
    def cancel!
      return unless approved?
      request, response = gateway.cancel self
      _cancel! if response.success?
      persist_transaction request, response
      true
    rescue TransmissionError => e
      handle_error 'cancel', e
    end

    # Refund payment that is settled
    def refund!

      puts "refund! #{status}"

      return unless completed?
      request, response = gateway.refund self
      _refund! if response.success?
      persist_transaction request, response
      true
    rescue TransmissionError => e
      handle_error 'refund', e
    end

    private

    def handle_error(request, error)
      transactions.create! \
        request: request,
        error:   error.serialize
    end

    def gateway
      SimplePayment::gateway
    end

    def persist_transaction(request, response)
      transactions.create! \
        request: request.serialize,
        response: response.serialize
    end
  end
end
